@@ -1,5 +1,5 @@
 /* Implementation of the STAT and FSTAT intrinsics.
-   Copyright (C) 2004 Free Software Foundation, Inc.
+   Copyright (C) 2004, 2005, 2006, 2007, 2009 Free Software Foundation, Inc.
    Contributed by Steven G. Kargl <kargls@comcast.net>.
 
 This file is part of the GNU Fortran 95 runtime library (libgfortran).
@@ -7,33 +7,26 @@ This file is part of the GNU Fortran 95 runtime library (libgfortran).
 Libgfortran is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public
 License as published by the Free Software Foundation; either
-version 2 of the License, or (at your option) any later version.
-
-In addition to the permissions in the GNU General Public License, the
-Free Software Foundation gives you unlimited permission to link the
-compiled version of this file into combinations with other programs,
-and to distribute those combinations without any restriction coming
-from the use of this file.  (The General Public License restrictions
-do apply in other respects; for example, they cover modification of
-the file, and distribution when not linked into a combine
-executable.)
+version 3 of the License, or (at your option) any later version.
 
 Libgfortran is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
-You should have received a copy of the GNU General Public
-License along with libgfortran; see the file COPYING.  If not,
-write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
-Boston, MA 02111-1307, USA.  */
+Under Section 7 of GPL version 3, you are granted additional
+permissions described in the GCC Runtime Library Exception, version
+3.1, as published by the Free Software Foundation.
+
+You should have received a copy of the GNU General Public License and
+a copy of the GCC Runtime Library Exception along with this program;
+see the files COPYING3 and COPYING.RUNTIME respectively.  If not, see
+<http://www.gnu.org/licenses/>.  */
 
-#include "config.h"
 #include "libgfortran.h"
 
-#ifdef HAVE_SYS_TYPES_H
-#include <sys/types.h>
-#endif
+#include <string.h>
+#include <errno.h>
 
 #ifdef HAVE_SYS_STAT_H
 #include <sys/stat.h>
@@ -43,13 +36,8 @@ Boston, MA 02111-1307, USA.  */
 #include <stdlib.h>
 #endif
 
-#ifdef HAVE_STRING_H
-#include <string.h>
-#endif
 
-#include <errno.h>
-
-#include "../io/io.h"
+#ifdef HAVE_STAT
 
 /* SUBROUTINE STAT(FILE, SARRAY, STATUS)
    CHARACTER(len=*), INTENT(IN) :: FILE
@@ -61,20 +49,18 @@ Boston, MA 02111-1307, USA.  */
    CHARACTER(len=*), INTENT(IN) :: FILE
    INTEGER, INTENT(OUT), :: SARRAY(13)  */
 
-extern void stat_i4_sub (char *, gfc_array_i4 *, GFC_INTEGER_4 *,
-			 gfc_charlen_type);
-iexport_proto(stat_i4_sub);
+/*extern void stat_i4_sub_0 (char *, gfc_array_i4 *, GFC_INTEGER_4 *,
+			   gfc_charlen_type, int);
+internal_proto(stat_i4_sub_0);*/
 
-void
-stat_i4_sub (char *name, gfc_array_i4 *sarray, GFC_INTEGER_4 *status,
-	     gfc_charlen_type name_len)
+static void
+stat_i4_sub_0 (char *name, gfc_array_i4 *sarray, GFC_INTEGER_4 *status,
+	       gfc_charlen_type name_len, int is_lstat __attribute__ ((unused)))
 {
   int val;
   char *str;
   struct stat sb;
 
-  index_type stride[GFC_MAX_DIMENSIONS];
-      
   /* If the rank of the array is not 1, abort.  */
   if (GFC_DESCRIPTOR_RANK (sarray) != 1)
     runtime_error ("Array rank of SARRAY is not 1.");
@@ -83,9 +69,6 @@ stat_i4_sub (char *name, gfc_array_i4 *sarray, GFC_INTEGER_4 *status,
   if (sarray->dim[0].ubound + 1 - sarray->dim[0].lbound < 13)
     runtime_error ("Array size of SARRAY is too small.");
 
-  if (sarray->dim[0].stride == 0)
-    sarray->dim[0].stride = 1;
-
   /* Trim trailing spaces from name.  */
   while (name_len > 0 && name[name_len - 1] == ' ')
     name_len--;
@@ -93,11 +76,17 @@ stat_i4_sub (char *name, gfc_array_i4 *sarray, GFC_INTEGER_4 *status,
   /* Make a null terminated copy of the string.  */
   str = gfc_alloca (name_len + 1);
   memcpy (str, name, name_len);
-  str[name_len] = '\0'; 
+  str[name_len] = '\0';
 
-  val = stat(str, &sb);
+  /* On platforms that don't provide lstat(), we use stat() instead.  */
+#ifdef HAVE_LSTAT
+  if (is_lstat)
+    val = lstat(str, &sb);
+  else
+#endif
+    val = stat(str, &sb);
 
-  if (val == 0) 
+  if (val == 0)
     {
       /* Device ID  */
       sarray->data[0 * sarray->dim[0].stride] = sb.st_dev;
@@ -116,7 +105,7 @@ stat_i4_sub (char *name, gfc_array_i4 *sarray, GFC_INTEGER_4 *status,
 
       /* Owner's gid  */
       sarray->data[5 * sarray->dim[0].stride] = sb.st_gid;
-  
+
       /* ID of device containing directory entry for file (0 if not available) */
 #if HAVE_STRUCT_STAT_ST_RDEV
       sarray->data[6 * sarray->dim[0].stride] = sb.st_rdev;
@@ -151,25 +140,46 @@ stat_i4_sub (char *name, gfc_array_i4 *sarray, GFC_INTEGER_4 *status,
 #endif
     }
 
-  if (status != NULL) 
+  if (status != NULL)
     *status = (val == 0) ? 0 : errno;
 }
+
+
+extern void stat_i4_sub (char *, gfc_array_i4 *, GFC_INTEGER_4 *,
+			 gfc_charlen_type);
+iexport_proto(stat_i4_sub);
+
+void
+stat_i4_sub (char *name, gfc_array_i4 *sarray, GFC_INTEGER_4 *status,
+	     gfc_charlen_type name_len)
+{
+  stat_i4_sub_0 (name, sarray, status, name_len, 0);
+}
 iexport(stat_i4_sub);
 
-extern void stat_i8_sub (char *, gfc_array_i8 *, GFC_INTEGER_8 *,
+
+extern void lstat_i4_sub (char *, gfc_array_i4 *, GFC_INTEGER_4 *,
 			 gfc_charlen_type);
-iexport_proto(stat_i8_sub);
+iexport_proto(lstat_i4_sub);
 
 void
-stat_i8_sub (char *name, gfc_array_i8 *sarray, GFC_INTEGER_8 *status,
+lstat_i4_sub (char *name, gfc_array_i4 *sarray, GFC_INTEGER_4 *status,
 	     gfc_charlen_type name_len)
+{
+  stat_i4_sub_0 (name, sarray, status, name_len, 1);
+}
+iexport(lstat_i4_sub);
+
+
+
+static void
+stat_i8_sub_0 (char *name, gfc_array_i8 *sarray, GFC_INTEGER_8 *status,
+	       gfc_charlen_type name_len, int is_lstat __attribute__ ((unused)))
 {
   int val;
   char *str;
   struct stat sb;
 
-  index_type stride[GFC_MAX_DIMENSIONS];
-      
   /* If the rank of the array is not 1, abort.  */
   if (GFC_DESCRIPTOR_RANK (sarray) != 1)
     runtime_error ("Array rank of SARRAY is not 1.");
@@ -178,9 +188,6 @@ stat_i8_sub (char *name, gfc_array_i8 *sarray, GFC_INTEGER_8 *status,
   if (sarray->dim[0].ubound + 1 - sarray->dim[0].lbound < 13)
     runtime_error ("Array size of SARRAY is too small.");
 
-  if (sarray->dim[0].stride == 0)
-    sarray->dim[0].stride = 1;
-
   /* Trim trailing spaces from name.  */
   while (name_len > 0 && name[name_len - 1] == ' ')
     name_len--;
@@ -188,9 +195,15 @@ stat_i8_sub (char *name, gfc_array_i8 *sarray, GFC_INTEGER_8 *status,
   /* Make a null terminated copy of the string.  */
   str = gfc_alloca (name_len + 1);
   memcpy (str, name, name_len);
-  str[name_len] = '\0'; 
+  str[name_len] = '\0';
 
-  val = stat(str, &sb);
+  /* On platforms that don't provide lstat(), we use stat() instead.  */
+#ifdef HAVE_LSTAT
+  if (is_lstat)
+    val = lstat(str, &sb);
+  else
+#endif
+    val = stat(str, &sb);
 
   if (val == 0)
     {
@@ -205,13 +218,13 @@ stat_i8_sub (char *name, gfc_array_i8 *sarray, GFC_INTEGER_8 *status,
 
       /* Number of (hard) links  */
       sarray->data[3 * sarray->dim[0].stride] = sb.st_nlink;
-  
+
       /* Owner's uid  */
       sarray->data[4 * sarray->dim[0].stride] = sb.st_uid;
 
       /* Owner's gid  */
       sarray->data[5 * sarray->dim[0].stride] = sb.st_gid;
-  
+
       /* ID of device containing directory entry for file (0 if not available) */
 #if HAVE_STRUCT_STAT_ST_RDEV
       sarray->data[6 * sarray->dim[0].stride] = sb.st_rdev;
@@ -246,11 +259,39 @@ stat_i8_sub (char *name, gfc_array_i8 *sarray, GFC_INTEGER_8 *status,
 #endif
     }
 
-  if (status != NULL) 
+  if (status != NULL)
     *status = (val == 0) ? 0 : errno;
 }
+
+
+extern void stat_i8_sub (char *, gfc_array_i8 *, GFC_INTEGER_8 *,
+			 gfc_charlen_type);
+iexport_proto(stat_i8_sub);
+
+void
+stat_i8_sub (char *name, gfc_array_i8 *sarray, GFC_INTEGER_8 *status,
+	     gfc_charlen_type name_len)
+{
+  stat_i8_sub_0 (name, sarray, status, name_len, 0);
+}
+
 iexport(stat_i8_sub);
 
+
+extern void lstat_i8_sub (char *, gfc_array_i8 *, GFC_INTEGER_8 *,
+			 gfc_charlen_type);
+iexport_proto(lstat_i8_sub);
+
+void
+lstat_i8_sub (char *name, gfc_array_i8 *sarray, GFC_INTEGER_8 *status,
+	     gfc_charlen_type name_len)
+{
+  stat_i8_sub_0 (name, sarray, status, name_len, 1);
+}
+
+iexport(lstat_i8_sub);
+
+
 extern GFC_INTEGER_4 stat_i4 (char *, gfc_array_i4 *, gfc_charlen_type);
 export_proto(stat_i4);
 
@@ -274,14 +315,51 @@ stat_i8 (char *name, gfc_array_i8 *sarray, gfc_charlen_type name_len)
 }
 
 
+/* SUBROUTINE LSTAT(FILE, SARRAY, STATUS)
+   CHARACTER(len=*), INTENT(IN) :: FILE
+   INTEGER, INTENT(OUT), :: SARRAY(13)
+   INTEGER, INTENT(OUT), OPTIONAL :: STATUS
+
+   FUNCTION LSTAT(FILE, SARRAY)
+   INTEGER LSTAT
+   CHARACTER(len=*), INTENT(IN) :: FILE
+   INTEGER, INTENT(OUT), :: SARRAY(13)  */
+
+extern GFC_INTEGER_4 lstat_i4 (char *, gfc_array_i4 *, gfc_charlen_type);
+export_proto(lstat_i4);
+
+GFC_INTEGER_4
+lstat_i4 (char *name, gfc_array_i4 *sarray, gfc_charlen_type name_len)
+{
+  GFC_INTEGER_4 val;
+  lstat_i4_sub (name, sarray, &val, name_len);
+  return val;
+}
+
+extern GFC_INTEGER_8 lstat_i8 (char *, gfc_array_i8 *, gfc_charlen_type);
+export_proto(lstat_i8);
+
+GFC_INTEGER_8
+lstat_i8 (char *name, gfc_array_i8 *sarray, gfc_charlen_type name_len)
+{
+  GFC_INTEGER_8 val;
+  lstat_i8_sub (name, sarray, &val, name_len);
+  return val;
+}
+
+#endif
+
+
+#ifdef HAVE_FSTAT
+
 /* SUBROUTINE FSTAT(UNIT, SARRAY, STATUS)
-   INTEGER, INTENT(IN) :: UNIT 
+   INTEGER, INTENT(IN) :: UNIT
    INTEGER, INTENT(OUT) :: SARRAY(13)
-   INTEGER, INTENT(OUT), OPTIONAL :: STATUS 
+   INTEGER, INTENT(OUT), OPTIONAL :: STATUS
 
    FUNCTION FSTAT(UNIT, SARRAY)
    INTEGER FSTAT
-   INTEGER, INTENT(IN) :: UNIT 
+   INTEGER, INTENT(IN) :: UNIT
    INTEGER, INTENT(OUT) :: SARRAY(13)  */
 
 extern void fstat_i4_sub (GFC_INTEGER_4 *, gfc_array_i4 *, GFC_INTEGER_4 *);
@@ -293,8 +371,6 @@ fstat_i4_sub (GFC_INTEGER_4 *unit, gfc_array_i4 *sarray, GFC_INTEGER_4 *status)
   int val;
   struct stat sb;
 
-  index_type stride[GFC_MAX_DIMENSIONS];
-      
   /* If the rank of the array is not 1, abort.  */
   if (GFC_DESCRIPTOR_RANK (sarray) != 1)
     runtime_error ("Array rank of SARRAY is not 1.");
@@ -303,9 +379,6 @@ fstat_i4_sub (GFC_INTEGER_4 *unit, gfc_array_i4 *sarray, GFC_INTEGER_4 *status)
   if (sarray->dim[0].ubound + 1 - sarray->dim[0].lbound < 13)
     runtime_error ("Array size of SARRAY is too small.");
 
-  if (sarray->dim[0].stride == 0)
-    sarray->dim[0].stride = 1;
-
   /* Convert Fortran unit number to C file descriptor.  */
   val = unit_to_fd (*unit);
   if (val >= 0)
@@ -318,7 +391,7 @@ fstat_i4_sub (GFC_INTEGER_4 *unit, gfc_array_i4 *sarray, GFC_INTEGER_4 *status)
 
       /* Inode number  */
       sarray->data[1 * sarray->dim[0].stride] = sb.st_ino;
-  
+
       /* File mode  */
       sarray->data[2 * sarray->dim[0].stride] = sb.st_mode;
 
@@ -330,7 +403,7 @@ fstat_i4_sub (GFC_INTEGER_4 *unit, gfc_array_i4 *sarray, GFC_INTEGER_4 *status)
 
       /* Owner's gid  */
       sarray->data[5 * sarray->dim[0].stride] = sb.st_gid;
-  
+
       /* ID of device containing directory entry for file (0 if not available) */
 #if HAVE_STRUCT_STAT_ST_RDEV
       sarray->data[6 * sarray->dim[0].stride] = sb.st_rdev;
@@ -365,7 +438,7 @@ fstat_i4_sub (GFC_INTEGER_4 *unit, gfc_array_i4 *sarray, GFC_INTEGER_4 *status)
 #endif
     }
 
-  if (status != NULL) 
+  if (status != NULL)
     *status = (val == 0) ? 0 : errno;
 }
 iexport(fstat_i4_sub);
@@ -379,8 +452,6 @@ fstat_i8_sub (GFC_INTEGER_8 *unit, gfc_array_i8 *sarray, GFC_INTEGER_8 *status)
   int val;
   struct stat sb;
 
-  index_type stride[GFC_MAX_DIMENSIONS];
-      
   /* If the rank of the array is not 1, abort.  */
   if (GFC_DESCRIPTOR_RANK (sarray) != 1)
     runtime_error ("Array rank of SARRAY is not 1.");
@@ -389,9 +460,6 @@ fstat_i8_sub (GFC_INTEGER_8 *unit, gfc_array_i8 *sarray, GFC_INTEGER_8 *status)
   if (sarray->dim[0].ubound + 1 - sarray->dim[0].lbound < 13)
     runtime_error ("Array size of SARRAY is too small.");
 
-  if (sarray->dim[0].stride == 0)
-    sarray->dim[0].stride = 1;
-
   /* Convert Fortran unit number to C file descriptor.  */
   val = unit_to_fd ((int) *unit);
   if (val >= 0)
@@ -416,7 +484,7 @@ fstat_i8_sub (GFC_INTEGER_8 *unit, gfc_array_i8 *sarray, GFC_INTEGER_8 *status)
 
       /* Owner's gid  */
       sarray->data[5 * sarray->dim[0].stride] = sb.st_gid;
-  
+
       /* ID of device containing directory entry for file (0 if not available) */
 #if HAVE_STRUCT_STAT_ST_RDEV
       sarray->data[6 * sarray->dim[0].stride] = sb.st_rdev;
@@ -451,7 +519,7 @@ fstat_i8_sub (GFC_INTEGER_8 *unit, gfc_array_i8 *sarray, GFC_INTEGER_8 *status)
 #endif
     }
 
-  if (status != NULL) 
+  if (status != NULL)
     *status = (val == 0) ? 0 : errno;
 }
 iexport(fstat_i8_sub);
@@ -477,3 +545,5 @@ fstat_i8 (GFC_INTEGER_8 *unit, gfc_array_i8 *sarray)
   fstat_i8_sub (unit, sarray, &val);
   return val;
 }
+
+#endif
